#Include This Once

Sub NewCodeGenAmd64( obj As iCodeGen, sendOutputTo As iOutput )
	obj = Class "cCodeGenAmd64"
	If Not IsInterface( obj, iCodeGen ) Then
		MyPrint "Failed to create iCodeGen"
		End
	End If
	obj.Output = sendOutputTo
End Sub

Class cCodeGenAmd64
	Instance nextLabelId_ As Long
	Instance output_ As iOutput

	Class Method Create()
		nextLabelId_ = 1
	End Method

	Interface iCodeGen
		Inherit IUnknown

		Property Get Output() As iOutput
			Property = output_
		End Property
		Property Set Output( sendOutputTo As iOutput )
			output_ = sendOutputTo
		End Property
		Method IncludeLine( value As String )
			output_.WriteLine( value )
		End Method
		Method Emit( value As String )
			' output_ a string with tab
			output_.Write( $Tab )
			output_.Write( value )
		End Method
		Method EmitLine( value As String )
			' output_ a string with tab and crlf
			output_.Write( $Tab )
			output_.WriteLine( value )
		End Method
		Method Allocate( ident As String, value As String )
			' Allocate storage for a static variable
			output_.Write( ident )
			output_.Write( "_" )
			output_.Write( $Tab )
			output_.Write( "dq" )
			output_.Write( $Tab )
			output_.WriteLine( value )
		End Method
		Method PostLabel( label As String )
			' Post a label to output_
			output_.Write( Label )
			output_.WriteLine( ":" )
		End Method
		Method NewLabel() As String
			' Generate a unique label
			Local value As String
			value = Format$( nextLabelId_, "0" )
			nextLabelId_ += 1
			Method = "L" + value
		End Method
		Method Clear()
			' Clear the primary register
			Me.EmitLine( "xor rax, rax" )
		End Method
		Method Negate()
			' Negate the primary register
			Me.EmitLine( "neg rax" )
		End Method
		Method NotPrimary()
			' Complement the primary register
			Me.EmitLine( "not rax" )
		End Method
		Method LoadConst( value As String )
			' Load a constant value to primary register
			Me.EmitLine( "mov rax, " + value )
		End Method
		Method LoadVar( variable As String )
			' Load a variable to primary register
			If Not symbols.Contains( variable ) Then UndefinedError( variable )
			Me.EmitLine( "mov rax, " + variable + "_" )
		End Method
		Method Push()
			' Push primary onto stack
			Me.EmitLine( "push rax" )
		End Method
		Method PopAdd()
			' Add top of stack to primary
			Me.EmitLine( "pop rbx" )
			Me.EmitLine( "add rax, rbx" )
		End Method
		Method PopSub()
			' Subtract top of stack to primary
			Me.EmitLine( "pop rbx" )
			Me.EmitLine( "xchg rax, rbx" )
			Me.EmitLine( "sub rax, rbx" )
		End Method
		Method PopMethod()
			' Methodtract primary from top of stack
			Me.EmitLine( "pop rbx" )
			Me.EmitLine( "xchg rax, rbx" )
			Me.EmitLine( "Method rax, rbx" )
		End Method
		Method PopMul()
			' Multiply top of stack by primary
			Me.EmitLine( "pop rbx" )
			Me.EmitLine( "imul rbx" )
		End Method
		Method PopDiv()
			' Divide top of stack by primary
			Me.EmitLine( "pop rbx" )
			Me.EmitLine( "xchg rax, rbx" )
			Me.EmitLine( "imul rbx" )
		End Method
		Method PopAnd()
			' And top of stack with primary
			Me.EmitLine( "pop rbx" )
			Me.EmitLine( "and rax, rbx" )
		End Method
		Method PopOr()
			' Or top of stack with primary
			Me.EmitLine( "pop rbx" )
			Me.EmitLine( "or rax, rbx" )
		End Method
		Method PopXor()
			' Xor top of stack with primary
			Me.EmitLine( "pop rbx" )
			Me.EmitLine( "xor rax, rbx" )
		End Method
		Method PopCompare()
			' Compare top of stack with primary
			Me.EmitLine( "pop rbx" )
			Me.EmitLine( "cmp rbx, rax" )
		End Method
		Method SetEqual()
			' Set d0 if compare was =
			Me.EmitLine( "mov rax,0" )
			Me.EmitLine( "setne al" ) ' Do the opposite ( 1->0, 0->-1 )
			Me.EmitLine( "dec rax" )  '
		End Method
		Method SetNEqual()
			' Set d0 if compare was !=
			Me.EmitLine( "mov rax,0" )
			Me.EmitLine( "sete al" )
			Me.EmitLine( "dec rax" )
		End Method
		Method SetGreater()
			' Set d0 if compare was >
			Me.EmitLine( "mov rax,0" )
			Me.EmitLine( "setle al" )
			Me.EmitLine( "dec rax" )
		End Method
		Method SetLess()
			' Set d0 if compare was <
			Me.EmitLine( "mov rax,0" )
			Me.EmitLine( "setge al" )
			Me.EmitLine( "dec rax" )
		End Method
		Method SetLessOrEqual()
			' Set d0 if compare was <=
			Me.EmitLine( "mov rax,0" )
			Me.EmitLine( "setg al" )
			Me.EmitLine( "dec rax" )
		End Method
		Method SetGreaterOrEqual()
			' Set d0 if compare was >=
			Me.EmitLine( "mov rax,0" )
			Me.EmitLine( "setl al" )
			Me.EmitLine( "dec rax" )
		End Method
		Method Store( variable As String )
			' Store primary to variable
			Me.EmitLine( "mov " + variable + "_, rax" )
		End Method
		Method Branch( label As String )
			' Branch unconditional
			Me.EmitLine( "jmp " + Label )
		End Method
		Method BranchFalse( label As String )
			' Branch false
			Me.EmitLine( "or rax, rax" )
			Me.EmitLine( "jz " + Label )
		End Method
		Method EndOfCode()
			Me.EmitLine( "ret" )
		End Method
		Method ReadConsole( variable As String )
			' Input to variable
			Me.EmitLine( "invoke @con_in, offset @inputbuf, 80" )
			Me.EmitLine( "invoke @atoi, offset @inputbuf" )
			Me.Store( variable )
		End Method
		Method WriteConsole()
			' Write from primary register
			Me.EmitLine( "invoke @itoa, rax, offset @inputbuf, 20" )
			Me.EmitLine( "invoke @con_show, offset @inputbuf" )
		End Method
		Method WriteConsoleConst( value As String )
			Me.EmitLine( "invoke @con_show, offset " + value )
		End Method
		Method Header()
			' Write header info
		End Method
		Method CreateVariables( list As iSymbols )
			Local sym As String
			Local keys() As String
			Local i, kc As Long
			sym = list.Symbols()
			kc = ParseCount( sym, "," )
			ReDim keys( kc-1 )
			Parse sym, keys(), ","

			For i=0 To kc-1
				Me.Allocate( keys( i ), "0" )
			Next
			Me.IncludeLine( "" )
		End Method
		Method Prologue1()
			' Write the prolog
			output_.Write( $Tab )
			output_.WriteLine( Resource$( RcData, %PROLOGUE1 ) )
		End Method
		Method Prologue2()
			output_.WriteLine( Resource$( RcData, %PROLOGUE2 ) )
		End Method
		Method Epilogue()
			' Write the epilog
			output_.WriteLine( Resource$( RcData, %EPILOGUE ) )
		End Method
	End Interface
End Class
