#Include This Once

' String to long map

Sub NewStringLongMap( ByRef obj As iStringLongMap )
	obj = Class "cStringLongMap"
	If Not IsInterface( obj, iStringLongMap ) Then
		MyPrint "Failed to create StringLongMap"
		End
	End If
End Sub

Class cStringLongMap
	Instance bag_ As iBag
	Class Method Create()
		Call NewBag( bag_ )
	End Method

	Interface iStringLongMap
		Inherit IUnknown

		' Returns index of added item
		Method Add( ByVal pKey As String, ByVal pItem As Long ) As Long
			Dim be As iBagEntry
			Call NewStringLongMapEntry( be, pKey, pItem )
			Method = bag_.Add( be )
		End Method

		' Empty the bag
		Method Clear()
			bag_.Clear()
		End Method

		' -1 if found, 0 if not found
		Property Get Contains( ByRef pKey As String ) As Long
			Dim be As iBagEntry
			Call NewStringLongMapEntry( be, pKey, 0 )

			Property = bag_.Contains( be )
		End Property

		' Number of items in the bag
		Property Get Count() As Long
			Property = bag_.Count
		End Property


		' Index of item if found, -1 if not found
		Method IndexOf( ByRef pKey As String ) As Long
			Dim be As iBagEntry
			Call NewStringLongMapEntry( be, pKey, 0 )
			Method = bag_.IndexOf( be )
		End Method

		Method Remove( ByRef pKey As String )
			Dim be As iBagEntry
			Call NewStringLongMapEntry( be, pKey, 0 )
			bag_.Remove( be )
		End Method

		Method RemoveAt( ByVal i As Long )
			bag_.RemoveAt( i )
		End Method

		' Get an item by index, returns empty string if out of bounds
		Property Get Value( ByVal pKey As String ) As Long
			Dim be As iBagEntry
			Dim iv As iStringLongPair
			Local idx As Long
			Property = 0
			idx = Me.IndexOf( pKey )
			be = bag_.Item( idx )
			iv = be
			If IsObject( iv ) Then
				Property = iv.Value
			End If
		End Property
		Property Get Key( ByVal index As Long ) As String
			Dim be As iBagEntry
			Dim iv As iStringLongPair
			Property = ""
			be = bag_.Item( index )
			iv = be
			If IsObject( iv ) Then
				Property = iv.Key
			End If
		End Property
	End Interface
End Class

Sub NewStringLongMapEntry( ByRef be As iBagEntry, ByRef newkey As String, ByRef value As Long )
	Local b As iStringLongPair

	be = Class "cStringLongMapEntry"
	If Not IsInterface( be, iBagEntry ) Then
		MyPrint "Failed to create iBagEntry for " + newkey
		End
	End If

	b = be
	If Not IsInterface( b, iStringLongPair ) Then
		MyPrint "Failed to create iStringLongPair for " + newkey
		End
	End If
	b.Key = newkey
	b.Value = value
End Sub

'=================================================================
' StringLongMap entries ( internal implementation class )
'=================================================================
Class cStringLongMapEntry
	Instance hashCode_ As Long
	Instance next_ As Long
	Instance key_ As String
	Instance value_ As Long

	Interface iStringLongPair
		Inherit IUnknown

		Property Get Value() As Long
			Property = value_
		End Property
		Property Set Value( ByVal value As Long )
			value_ = value
		End Property

		Property Get Key () As String
			Property = key_
		End Property
		Property Set Key ( ByVal pVal  As String )
			Local p As Byte Ptr
			Local i As Long

			If key_ <> pVal Then
				key_  = pVal

				hashCode_ = 1

				p = StrPtr( key_ )
				While @p
					hashCode_ = 1610612741 * hashCode_ + @p
					Incr p
				Wend
				hashCode_ And= &H7FFFFFFF
			End If
		End Property
	End Interface

	Interface iBagEntry
		Inherit IUnknown

		Method Same( ByVal value As iBagEntry ) As Long
			Local a As iStringLongPair
			Local b As iStringLongPair
			a = Me
			b = value
			If IsObject( a ) And IsObject( b ) Then
				Method = IIf&( LCase$( a.Key ) = LCase$( b.Key ), 1, 0 )
			End If
		End Method

		Property Get HashCode () As Long
			Property = hashCode_
		End Property
		Property Set HashCode ( ByVal pHashCode  As Long )
			hashCode_  = pHashCode
		End Property

		Property Get Next () As Long
			Property = next_
		End Property
		Property Set Next ( ByVal pNext  As Long )
			next_  = pNext
		End Property
	End Interface
End Class
